/* PIC16C84 programmer, using PROG84 hardware
 *
 *  Copyright (C) 2002 Shevek <gpprog@anarres.org>
 *
 * With reference to code by:
 * Steve Marchant (Steve.Marchant@Nottingham.ac.uk)
 * Laszlo Nemeth (nemeth@eik.bme.hu)
 * Brian C. Lane (brian@nexuscomputing.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/resource.h>
#include <sched.h>
#include <sys/timeb.h>
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/io.h>
#include <time.h>
#include <sys/time.h>
#include <stdarg.h>

#define CMD_LOAD_CONFIG		0x00
#define CMD_MYSTERY_01		0x01	/* Part of unprotect sequence */
#define CMD_LOAD_PROGRAM	0x02
#define CMD_LOAD_DATA		0x03
#define CMD_READ_PROGRAM	0x04
#define CMD_READ_DATA		0x05
#define CMD_INCR_ADDR		0x06
#define CMD_MYSTERY_07		0x07	/* Part of unprotect sequence */
#define CMD_BEGIN_PROG		0x08
#define CMD_WIPE_PROGRAM	0x09
#define CMD_WIPE_DATA		0x0b
#define CMD_END_PROG		0x0e

static char *pic_cmds[] = {
	[CMD_LOAD_CONFIG]	= "Load Configuration",
	[CMD_LOAD_PROGRAM]	= "Load Data for Program Memory",
	[CMD_READ_PROGRAM]	= "Read Data from Program Memory",
	[CMD_INCR_ADDR]		= "Increment Address",
	[CMD_BEGIN_PROG]	= "Begin Programming",
	[CMD_LOAD_DATA]		= "Load Data for Data Memory",
	[CMD_READ_DATA]		= "Read Data from Data Memory",
	[CMD_WIPE_PROGRAM]	= "Bulk erase program memory",
	[CMD_WIPE_DATA]		= "Bulk erase data memory",
	[CMD_END_PROG]		= "End programming",

	[CMD_MYSTERY_01]	= "Mystery unprotect command 1",
	[CMD_MYSTERY_07]	= "Mystery unprotect command 7",
};

	/* See page 3-63, ICSP guide */
#define PIC_PROG_BASE	0x0000
#define PIC_PROG_SIZE	0x0400
#define PIC_PROG_BASE	0x0000
#define PIC_PROG_END	(PIC_PROG_BASE + PIC_PROG_SIZE)
#define PIC_CONF_BASE	0x2000
#define PIC_CONF_SIZE	0x0004
#define PIC_CONF_END	(PIC_CONF_BASE + PIC_CONF_SIZE)
#define PIC_DATA_BASE	0x2100
#define PIC_DATA_SIZE	0x0040
#define PIC_DATA_END	(PIC_DATA_BASE + PIC_DATA_SIZE)

#define PIC_IMGSIZE		0x4000

#define PIC_DL_BINARY	10
#define PIC_DL_DATA		8
#define PIC_DL_CMD		6
#define PIC_DL_PROGRAM	5
#define PIC_DL_ACTION	4
#define PIC_DL_MESSAGE	1

typedef unsigned short	port_addr_t;
typedef unsigned char	port_data_t;

typedef unsigned long	pic_addr_t;
typedef unsigned short	pic_insn_t;

typedef
struct _pic_byte_t {
	pic_insn_t	 insn;
	char		 used;
}
pic_byte_t;

typedef
struct _pic_code_t {
	pic_byte_t		 image[PIC_IMGSIZE];
}
pic_code_t;

#define pic_code_new()		((pic_code_t *)calloc(1,sizeof(pic_code_t)))
#define pic_code_free(x)	(free(x))

#define	true	1
#define	false	0

#define PORT_ADDR	0x378
#define PORT_DEV	"/dev/lp0"

static port_data_t	 pic_outdata;

static int			 pic_dlevel;

void
pic_printf(int dlevel, char *fmt, ...)
{
	va_list	 ap;

	va_start(ap, fmt);
	if (dlevel <= pic_dlevel)
		vprintf(fmt, ap);
	va_end(ap);
}

static unsigned long
pic_strtohex(const char *str, off_t off, size_t len)
{
	const char		*cp;
	unsigned long	 out;

	out = 0;
	for (cp = str + off; len; len--) {
		switch (*cp) {
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				out = (out << 4) + (*cp - '0');
				break;
			case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
				out = (out << 4) + (*cp - 'a' + 10);
				break;
			case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
				out = (out << 4) + (*cp - 'A' + 10);
				break;
			default:
				printf("Unrecognised hex char %c at %s\n", *cp, cp);
				exit(1);
		}
		cp++;
	}

	return out;
}

static inline char
pic_hexchar(unsigned char x)
{
	if (x < 10)
		return x + '0';
	else if (x >= 16)
		return 'X';
	else
		return x - 10 + 'A';
}

static FILE *
pic_fopen(const char *path, char *mode)
{
	FILE	*fp;

	if (!path || !*path) {
		printf("No filename given to pic_fopen\n");
		exit(1);
	}

	if (strcmp(path, "-") == 0) {
		if (strcmp(mode, "r") == 0)
			return stdin;
		else
			return stdout;
	}

	seteuid(getuid());
	setegid(getgid());
	fp = fopen(path, mode);
	seteuid(0);
	setegid(0);

	if (fp == NULL) {
		printf("Failed to open '%s' mode %s\n", path, mode);
		perror("fopen");
	}

	return fp;
} 

static void
pic_fclose(FILE *stream)
{
	if (fileno(stream) > 2)
		fclose(stream);
}

static char *
pic_fgets(char *line, int size, FILE *f)
{
	while (fgets(line, size, f)) {
		line[strcspn(line, "\r\n")] = '\0';
		// if (line[0] != ':' || strlen(line) <= 11)
		if (line[0] != ':')
			continue;
		return line;
	}

	return NULL;
}

static const char *
pic_loc_desc(int addr)
{
	if (addr < 0x400)
		return ("Program memory");
	if (addr > 0x1fff && addr < 0x2004)
		return ("ID location");
	if (addr == 0x2007)
		return ("Config Word");
	if (addr > 0x20ff && addr < 0x2140)
		return ("EE Data memory");
	return NULL;
}

	/* High, Low, Tristate high impedance */

#define BIT_CLK_HI	(1<<1)
#define BIT_CLK_LO	(0)
#define BIT_CLK_HZ	(1<<5)

#define BIT_MCLR_HI	(1<<3)
#define BIT_MCLR_LO	(0)
#define BIT_MCLR_HZ	(1<<4)

#define BIT_DATA_HI	(1<<0)
#define BIT_DATA_LO	(0)
#define BIT_DATA_HZ	(1<<2)

static const char *
pic_bitprint(int x)
{
	static char	 buf[79][20];
	static int	 off = 0;

	off = (off + 1) % 20;

	sprintf(buf[off], "%c%c%c%c%c%c%c%c "
				"(CLK=%s,%c, DATA=%s,%c, MCLR=%s,%sV)",
		((x & (1 << 7)) ? 'X' : '_'),
		((x & (1 << 6)) ? 'X' : '_'),
		((x & (1 << 5)) ? 'X' : '_'),
		((x & (1 << 4)) ? 'X' : '_'),
		((x & (1 << 3)) ? 'X' : '_'),
		((x & (1 << 2)) ? 'X' : '_'),
		((x & (1 << 1)) ? 'X' : '_'),
		((x & (1 << 0)) ? 'X' : '_'),

		((x & (1 << 5)) ? "dis" : "en"),
		((x & (1 << 1)) ? '1' : '0'),
		((x & (1 << 2)) ? "dis" : "en"),
		((x & (1 << 0)) ? '1' : '0'),
		((x & (1 << 4)) ? "dis" : "en"),
		((x & (1 << 3)) ? "12" : "5")

		);

	return buf[off];
}

static int			 pic_fd = -1;
static port_addr_t	 pic_port_out = PORT_ADDR;
static port_addr_t	 pic_port_in = PORT_ADDR + 1;

static void
pic_outb(port_data_t val)
{
	pic_outdata = val;
	pic_printf(PIC_DL_BINARY, "<< %s\n", pic_bitprint(val));
	outb(val, pic_port_out);
}

static unsigned short
pic_inb()
{
	unsigned short	 val = inb(pic_port_in);
	pic_printf(PIC_DL_BINARY, ">> %s\n", pic_bitprint(val));
	return val;
}

static int
pic_port_close()
{
	struct sched_param schp;

	pic_printf(PIC_DL_ACTION, "Closing PIC port\n");
	pic_outb(0xe4);   /* 11100100 RB6=HZ RB7=HZ CLR=HI */
	ioperm(pic_port_out, 2, 0);
	close(pic_fd);
	pic_fd = -1;

	setpriority(PRIO_PROCESS, 0, 0);

	schp.sched_priority = SCHED_OTHER;
	sched_setparam(0, &schp);

	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);

	return true;
}

static void
pic_sighandler(int dummy)
{
	if (pic_fd >= 0)
		pic_port_close();
}

static int
pic_port_open()
{
	struct sched_param schp;

	if (pic_fd >= 0)
		pic_port_close();

	setpriority(PRIO_PROCESS, 0, -19);

	schp.sched_priority = SCHED_RR;
	sched_setparam(0, &schp);

	signal(SIGINT, pic_sighandler);
	signal(SIGTERM, pic_sighandler);

	pic_printf(PIC_DL_ACTION,
			"Attempting to use %s port 0x%x\n", PORT_DEV, pic_port_out);
	pic_fd = open (PORT_DEV, O_RDWR | O_NONBLOCK);
	if (pic_fd < 0) {
		printf("Port %s is locked.\n", PORT_DEV);
		perror("open");
		return false;
	}

	if (ioperm(pic_port_out, 2, 1) < 0) {
		perror("ioperm");
		return false;
	}

	return true;
}

static void
pic_usleep(unsigned int us)
{
	struct timeval	 now, wt;

	if (us > 1000) {
		/* We can probably do milliseconds */
		usleep(us);
		return;
	}

	gettimeofday(&now, NULL);

	wt.tv_usec = (now.tv_usec + us) % 1000000;
	wt.tv_sec = now.tv_sec + ((now.tv_usec + us) / 1000000);

	while (1) {
	    gettimeofday(&now, NULL);
		if (now.tv_sec > wt.tv_sec)
			return;
		if ((now.tv_sec == wt.tv_sec) &&
			 (now.tv_usec >= wt.tv_usec))
			return;
	}
}

	/* The 16F84 needs only 1us. The CMOS chips need more? */
#define pic_tick() pic_usleep(1)

static void
pic_reset(void)
{
	pic_printf(PIC_DL_ACTION, "Initialising PIC...\n");
	/* Exit & re-enter program mode to reset cp */
	pic_outb(0xe4);   /* 11100100 RB6=HZ RB7=HZ CLR=HI */
	pic_usleep(100);
	pic_outb(0xf4);   /* 11110100 RB6=HZ RB7=HZ CLR=LO */
	pic_usleep(100);
	pic_outb(0xd0);   /* 11010000 RB6=LO RB7=LO CLR=LO */
	pic_usleep(100);
	pic_outb(0xc8);   /* 11001000 RB6=LO RB7=LO CLR=12V */
	pic_usleep(100);
	pic_outb(0xcc);   /* 11001100 RB6=LO RB7=HZ CLR=12V */
	pic_usleep(100);
}

static void
pic_send_cmd(short d)
{
	short i, j, k;


	pic_printf(PIC_DL_CMD, "pic_send_cmd(%hd): %s\n",
			d,
			(pic_cmds[d] ? pic_cmds[d] : "Unknown"));

	j = d;
	pic_outb(0xcc);		/* CK=LO DAT=HZ */
	pic_tick();
	for (i = 0; i < 6; i++) {
		k = (j & 1) + 0xca;		/* CK=HI DAT=LSB */
		pic_outb(k);
		pic_tick();
		pic_outb(( k & 0xfd ));
		pic_tick();				/* CK=LO DAT=DAT */
		j >>= 1;
	}
	pic_outb(0xcc);		/* CK=LO DAT=HZ */
	pic_tick();
}

static void
pic_send_insn(pic_insn_t d)
{
	short i, k;
	long j;

	pic_printf(PIC_DL_DATA, "pic_send_insn(%4.4X)\n", d);

	j = (d & 0x3fff) * 2;
	pic_outb(0xcc);		/* CK=LO DAT=HZ */
	pic_tick();
	for (i = 0; i < 16; i++) {
		k = (j & 1) + 0xca;		/* CK=HI DAT=LSB */
		pic_outb(k);
		pic_tick();
		pic_outb((k & 0xfd));
		pic_tick();				/* CK=LO DAT=DAT */
		j >>= 1;
	}
	pic_outb(0xcc);		/* CK=LO DAT=HZ */
	pic_tick();
}

static pic_insn_t
pic_recv_insn()
{
	pic_insn_t	 insn;
	int			 i;
	int			 data;

	pic_printf(PIC_DL_DATA, "pic_recv_insn(BEGIN)\n");

	insn = 0;
	for (i = 0; i < 16; i++) {
		pic_outb(0xce);	/* CK=HI DAT=HZ */
		pic_tick();  
		data = pic_inb();
		pic_outb(0xcc);	/* CK=LO DAT=HZ */
		pic_tick();
		/* Read LSB first */
		if (data & 64)
			insn |= (1 << i);
	}
	insn = ((insn & 0x7ffe) >> 1);

	pic_printf(PIC_DL_DATA, "pic_recv_insn(%4.4x)\n", insn);

	return insn;
}

static pic_insn_t
pic_read_chip_program(void)
{
	pic_send_cmd(CMD_READ_PROGRAM);
	return pic_recv_insn();
}

static short
pic_read_chip_data(void)
{
	pic_send_cmd(CMD_READ_DATA);
	return pic_recv_insn();
}

static void
pic_write_chip_data(pic_insn_t insn)
{
	pic_send_cmd(CMD_LOAD_DATA);
	pic_send_insn(insn);

	pic_send_cmd(CMD_BEGIN_PROG);
	pic_usleep(10000);				/* Wait 10ms */
	pic_send_cmd(CMD_END_PROG);		/* Only for CMOS */
}

static int
pic_write_chip_program(pic_insn_t insn)
{
	pic_insn_t	 oinsn;

	oinsn = pic_read_chip_program();
	if (oinsn != insn) {
		pic_send_cmd(CMD_LOAD_PROGRAM);
		pic_send_insn(insn);

		pic_send_cmd(CMD_BEGIN_PROG);
		pic_usleep(10000);				/* Wait 10ms */
		// pic_send_cmd(CMD_END_PROG);		/* Only for CMOS */

		oinsn = pic_read_chip_program();
		if (insn != oinsn) {
			printf("ERROR: Wrote %4.4x, read %4.4x\n",
							insn, oinsn);
			pic_port_close();	/* It must be open. */
			exit(1);
			return false;
		}
	}

	return true;
}

	/* I have no idea what the operand to this instruction is. */
static void
pic_write_config(pic_insn_t insn)
{
	insn = (insn & 0x0f) | 0x10;		/* Disable code protection */
	pic_send_cmd(CMD_LOAD_CONFIG);
	pic_send_insn(insn);
}

static void
pic_unprotect()
{
	int	 i;

	/* MicroChip recommended initialisation procedure */

	pic_write_config(16L);	/* Code protect off */

	for (i = 0; i <= 6; i++)	/* Skip to 0x2007 */
		pic_send_cmd(CMD_INCR_ADDR);

	/* I don't know why. The book says so. */
	pic_send_cmd(CMD_MYSTERY_01);
	pic_send_cmd(CMD_MYSTERY_07);
	pic_send_cmd(CMD_BEGIN_PROG);
	pic_usleep(10000);
	pic_send_cmd(CMD_END_PROG);
	pic_send_cmd(CMD_MYSTERY_01);
	pic_send_cmd(CMD_MYSTERY_07);
}

static void
pic_wipe()
{
	pic_reset();
	/* If protected */
	// if (pic_read_chip_config() & 0x10)
	pic_unprotect();
	pic_reset();
	/* There should be a "Load data all 1s" here? Page 3-67 */
	pic_send_cmd(CMD_WIPE_PROGRAM);
	pic_send_cmd(CMD_BEGIN_PROG);
	pic_usleep(10000);
	pic_reset();
}

static short
pic_read_chip_config(void)
{
	short i;

	pic_reset();
	pic_write_config(0L);	/* Skip to 0x2000 */
	for (i = 0; i <= 6; i++)	/* Skip to 0x2007 */
		pic_send_cmd(CMD_INCR_ADDR);
	i = pic_read_chip_program() & 0x1f;
	return i;
}

static pic_code_t *
pic_read_file(const char *path)
{
	pic_code_t	*program;
	FILE		*hexfile;
	char		 line[256];
	short		 csum;		/* unsigned char */
	short		 datasum;	/* unsigned char */
	int			 i;

	int			 insns;
	int			 addr;
	pic_insn_t	 insn;
	int			 rtype;

	pic_printf(PIC_DL_ACTION, "Loading %s into memory...\n", path);

	if (!(hexfile = pic_fopen(path, "r")))
		return NULL;

	program = pic_code_new();

	while (pic_fgets(line, 256, hexfile)) {
  		csum = pic_strtohex(line, strlen(line) - 2, 2);

		datasum = 0;
		for(i = 1; i < (strlen(line) - 2); i += 2) {
			datasum = (datasum + pic_strtohex(line, i, 2));
		}
		datasum = (-datasum) & 0xff;

		if (datasum != csum) {
			/* XXX Print line number */
			printf("Checksum fails on line %s\n", line);
			free(program);
			program = NULL;
			break;
		}

		rtype = pic_strtohex(line, 7, 2);		/* Bytes 7-8 */

		if (rtype == 0x01) {
			pic_printf(PIC_DL_ACTION, "Read EOF record\n");
			break;	/* EOF record */
		}
		else if (rtype == 0x00) {
			/* Data record */
		}
		else {
			printf("Unrecognised record type %d\n", rtype);
			continue;
		}

				/* Convert byte count to insn count */
		insns = pic_strtohex(line, 1, 2) / 2;	/* Bytes 1-2 */
		addr = pic_strtohex(line, 3, 4) >> 1;	/* Bytes 3-6 */

		pic_printf(PIC_DL_PROGRAM,
				"0x%4.4x[%d]: %s\n", addr, insns, line + 9);

		if (strlen(line) < insns * 4 + 11) {
			printf("Short line: Should contain %d insns\n", insns);
			continue;
		}

		for (i = 0; i < insns; i++) {
			/* Essentially a swab() */
			insn =
					(pic_strtohex(line, (i * 4) + 11, 2) << 8) +
					pic_strtohex(line, (i * 4) + 9, 2);
			program->image[addr].used = true;
			program->image[addr].insn = insn;
			// printf("%4.4x: %d\n", addr, insn);
			addr++;		/* Where we are writing */
		}
	}

	fclose(hexfile);

	return program;
}

static pic_code_t *
pic_read_chip()
{
	pic_code_t	*program;
	pic_insn_t	 insn;
	pic_addr_t	 addr;

	if (!pic_port_open()) {
		printf("Failed to open port %d\n", pic_port_out);
		return false;
	}

	if ((pic_read_chip_config() & 0x10) == 0) {
		printf("Device is code protected, cannot read.\n");
		return NULL;
	}

	pic_reset();

	program = pic_code_new();

	for (addr = PIC_PROG_BASE; addr < PIC_PROG_END; addr++) {
		insn = pic_read_chip_program();

		program->image[addr].used = true;	/* XXX Not if it's 0x377 */
		program->image[addr].insn = insn;

		pic_send_cmd(CMD_INCR_ADDR);
	}

	for (addr = PIC_PROG_END - 1; addr >= PIC_PROG_BASE; addr--) {
		insn = program->image[addr].insn;
		if (insn == 0x000)
			program->image[addr].used = false;
		else if (insn == 0x3fff)
			program->image[addr].used = false;
		else
			break;
	}

	pic_reset();
	pic_write_config(0L);

	for (addr = PIC_CONF_BASE; addr < PIC_CONF_END; addr++) {
		insn = pic_read_chip_program() & 0xf;	/* XXX Check why */
		program->image[addr].used = true;
		program->image[addr].insn = insn & 0x0f;
		pic_send_cmd(CMD_INCR_ADDR); 
	}

	/* Skip to 2007 */
	pic_send_cmd(CMD_INCR_ADDR); 
	pic_send_cmd(CMD_INCR_ADDR); 
	pic_send_cmd(CMD_INCR_ADDR); 
	addr = 0x2007;

	/* This is the config data at 2007 */
	insn = pic_read_chip_program() & 0x1f;

	program->image[addr].used = true;
	program->image[addr].insn = insn;

	/* Why? */
	// pic_reset();
	// pic_write_config(0L);


#if 0
	for (i = 0; i <= 7; i++) {
		GetDeviceEEBlockString (l, i * 8);
		fprintf (hexfile, "%s\n", l);
	}
#endif

	pic_port_close();

	return program;
}

static int
pic_write_chip(pic_code_t *program)
{
	pic_insn_t	 insn;
	pic_addr_t	 addr;

	const char	*name;
	int			 ret;

	int			 i;

	ret = true;

	if (!pic_port_open()) {
		printf("Failed to open port %d\n", pic_port_out);
		return false;
	}

	pic_reset();

	for (addr = PIC_PROG_BASE; addr < PIC_PROG_END; addr++) {
		name = pic_loc_desc(addr);
		if (!name) {
			printf("Address %4.4lx is illegal - ignored\n", addr);
		}
		else if (program->image[addr].used) {
			insn = program->image[addr].insn;

			pic_printf(PIC_DL_PROGRAM, "%s Addr: %4.4lx := %4.4x\n",
							name, addr, insn);

			/* XXX Temp \n */
			if (!pic_write_chip_program(insn))
				ret = false;
		}

		pic_send_cmd(CMD_INCR_ADDR); 
	}

		/* Should this be program->image[0x2007].insn? */
	pic_write_config(0L);
	for (addr = 0x2000; addr < 0x2004; addr++) {
		insn = program->image[addr].insn;
		if (!pic_write_chip_program(insn))
			ret = false;
		pic_send_cmd(CMD_INCR_ADDR); 
	}

	/* Write the EE data? */

	/* Write location 2007 */
	addr = 0x2007;
	insn = program->image[addr].insn;
	insn = (insn & 0x0f) | 0x10;	/* Don't set CP bit */
	printf("Config word is %4.4x\n", insn);

	/* Actually, either of these techniques works. */
#if 0
	pic_send_cmd(CMD_INCR_ADDR); 
	pic_send_cmd(CMD_INCR_ADDR); 
	pic_send_cmd(CMD_INCR_ADDR); 
	pic_write_chip_program(insn);
#else
	/* Load the actual configuration */
	pic_reset();
	pic_write_config(insn);
	for (i = 0; i < 7; i++)
		pic_send_cmd(CMD_INCR_ADDR);
	pic_send_cmd(CMD_BEGIN_PROG);
	pic_usleep(10000);
	pic_send_cmd(CMD_END_PROG);
#endif

	pic_reset();

	pic_port_close();

	return ret;
}

static void
pic_write_file_line(pic_code_t *program, FILE *out, pic_addr_t addr)
{
	pic_insn_t		 insn;
	char			 line[256];
	char			*cp;
	int				 i;
	unsigned char	 lo;
	unsigned char	 hi;
	unsigned short	 datasum;

	sprintf(line, ":__%4.4lX00", (addr << 1));
	cp = line + strlen(line);

	/* This is essentially the two nibbles that get printed. */
	datasum = ((addr << 1) & 0xff) + (addr >> 7);

	for (i = 0; i < 0x08; i++) {
		if (!(program->image[addr].used)) {
			addr += (0x08 - i);
			break;
		}
		insn = program->image[addr].insn;
		lo = insn & 0xff;
		hi = (insn >> 8) & 0x3f;

		datasum = datasum + hi + lo;

		*cp++ = pic_hexchar((lo>>4) & 15);
		*cp++ = pic_hexchar(lo & 15);
		*cp++ = pic_hexchar((hi>>4) & 15);
		*cp++ = pic_hexchar(hi & 15);
		addr++;
	}

	if (!i)	/* An empty line */
		return;

	i = i * 2;	/* Convert insn count to byte count */
	datasum += i;
	line[1] = pic_hexchar((i>>4) & 15);
	line[2] = pic_hexchar(i & 15);

	/* Checksum */
	datasum = (-datasum) & 0xff;
	*cp++ = pic_hexchar((datasum>>4) & 15);
	*cp++ = pic_hexchar(datasum & 15);

	*cp = '\0';
	fprintf(out, "%s\n", line);
}

static int
pic_write_file(pic_code_t *program, const char *path)
{
	pic_addr_t	 addr;
	FILE		*hexfile;

	if (!(hexfile = pic_fopen(path, "w")))
		return false;

	/* Arse covering */
	program->image[0x2004].used = false;
	program->image[0x2008].used = false;

		/* addr += 0x08 */
	for (addr = PIC_PROG_BASE; addr < PIC_PROG_END; addr += 0x08) {
		pic_write_file_line(program, hexfile, addr);
	}

	for (addr = PIC_CONF_BASE; addr < PIC_CONF_END; addr += 0x08) {
		pic_write_file_line(program, hexfile, addr);
	}

	pic_write_file_line(program, hexfile, 0x2007);

	for (addr = PIC_DATA_BASE; addr < PIC_DATA_END; addr += 0x08) {
		pic_write_file_line(program, hexfile, addr);
	}

	fprintf(hexfile, ":00000001FF\n");	/* EOF */

	pic_fclose(hexfile);

	return true;
}

static int
pic_write_screen(pic_code_t *program)
{
	pic_insn_t	 insn;
	pic_addr_t	 addr;
	int			 i;

	i = 0;

#define LLEN 10

	for (addr = PIC_PROG_BASE; addr < PIC_PROG_END; addr++) {
		if (!program->image[addr].used) { i = 0; continue; }
		if (!(i++ % LLEN)) printf("\n%4.4lX: ", addr);
		insn = program->image[addr].insn;
		printf("%4.4X ", insn);
	}

	i = 0;

	for (addr = 0x2000; addr < 0x2004; addr++) {
		if (!(i++ % LLEN)) printf("\n%4.4lX: ", addr);
		insn = program->image[addr].insn;
		printf("%4.4X ", insn);
	}

	addr = 0x2007;
	insn = program->image[addr].insn;
	printf("\n%4.4lX: %4.4X  ", addr, insn);

	i = 0;

	for (addr = 0x2100; addr < 0x2140; addr++) {
		if (!program->image[addr].used) { i = 0; continue; }
		if (!(i++ % LLEN)) printf("\n%4.4lX: ", addr);
		insn = program->image[addr].insn;
		printf("%4.4X ", insn);
	}

	printf("\n");

	return true;
}

static int
pic_show_chip_config(void)
{
	unsigned short	 config;
	char			 tmp[PIC_CONF_END - PIC_PROG_BASE + 1];
	int				 i;

	if (!pic_port_open()) {
		printf("Failed to open port %d\n", pic_port_out);
		return false;
	}

	pic_reset();
	pic_write_config(0);	/* Skip to 0x2000 */

	tmp[4] = '\0';
	for (i = 0; i < 4; i++) {
		tmp[i] = pic_hexchar(pic_read_chip_program() & 0xf);
		pic_send_cmd(CMD_INCR_ADDR);
	}

	pic_send_cmd(CMD_INCR_ADDR);
	pic_send_cmd(CMD_INCR_ADDR);
	pic_send_cmd(CMD_INCR_ADDR);

	config = pic_read_chip_program() & 0x1f;
	printf("Config word:        %2.2x\n", config);
	printf("Config bits:        %c%c%c%c%c\n",
					((config & (1 << 4)) ? '1' : '0'),
					((config & (1 << 3)) ? '1' : '0'),
					((config & (1 << 2)) ? '1' : '0'),
					((config & (1 << 1)) ? '1' : '0'),
					((config & (1 << 0)) ? '1' : '0'));
	printf("Hex ID nibbles:     %s\n", tmp);

	printf("Code Protection:    O%s\n", (config & 0x10) ? "FF" : "N");
	printf("Power up timer:     O%s\n", (config & 0x08) ? "N" : "FF");
	printf("Watchdog timer:     O%s\n", (config & 0x04) ? "N" : "FF");

	printf("Oscillator type:    ");
	switch (config & 0x03) {
		case 0:
			printf("LP\n");
			break;
		case 1:
			printf("XT\n");
			break;
		case 2:
			printf("HS\n");
			break;
		case 3:
			printf("RC\n");
			break;
	}

	pic_port_close();

	return true;
}

static void
pic_usage(void)
{
	printf("\nUsage: gpprog [-cprw] [-v N] hexfile\n");
	printf("       -h     -- display this help screen\n");
	printf("       -c     -- display chip config\n");
	printf("       -p     -- program chip from file\n");
	printf("       -r     -- read chip into file\n");
	printf("       -w     -- wipe chip\n");
	printf("       -v     -- set verbosity (0-10)\n");
}

int
main(int argc, char *argv[])
{
	/* --------------- MAIN --------------- */

	char	 ch;

	char	*fname;
	int		 action;

	pic_code_t	*program;

	pic_dlevel = 1;

	action = 0;

#define ACT_CONFIG	0x01
#define ACT_PROG	0x02
#define ACT_READ	0x04

	while ((ch = getopt(argc, argv, "chprv:")) != EOF) {
		switch(ch) {
			case 'h':
				pic_usage();
				exit(0);
			case 'c':
				action = ACT_CONFIG;
				break;
			case 'p':
				action = ACT_PROG;
				break;
			case 'r':
				action = ACT_READ;
				break;
			case 'v':
				pic_dlevel = atoi(optarg);
				break;
			case '?':
			default:
				exit(0);
		}
	}
	argv += optind;
	argc -= optind;

	pic_printf(PIC_DL_MESSAGE,
		"GPPROG - PIC16[CF]84 ICSP by Shevek <gpprog@anarres.org>\n");

	if (argc) {
		fname = argv[0];
	}
	else {
		fname = NULL;
	}

	if (action & ACT_CONFIG) {
		pic_show_chip_config();
	}
	else if (action & ACT_READ) {
		program = pic_read_chip();
		pic_write_screen(program);
		pic_write_file(program, fname);
		pic_code_free(program);
	}
	else if (action & ACT_PROG) {
		program = pic_read_file(fname);
		pic_write_screen(program);
		pic_write_chip(program);
		pic_code_free(program);
		/* XXX Protect device if (conf & 0x10) */
	}
	else {
		pic_usage();
	}

	return 0;
}
